@@ -1,7 +1,7 @@
 package Complete::Util;
 
-our $DATE = '2014-07-13'; # DATE
-our $VERSION = '0.13'; # VERSION
+our $DATE = '2014-12-05'; # DATE
+our $VERSION = '0.14'; # VERSION
 
 use 5.010001;
 use strict;
@@ -22,6 +22,11 @@ our %SPEC;
 $SPEC{complete_array_elem} = {
     v => 1.1,
     summary => 'Complete from array',
+    description => <<'_',
+
+Will sort the resulting completion list, so you don't have to presort the array.
+
+_
     args => {
         array => { schema=>['array*'=>{of=>'str*'}], pos=>0, req=>1 },
         word  => { schema=>[str=>{default=>''}], pos=>1 },
@@ -110,6 +115,7 @@ Windows is supported, on Windows PATH will be split using /;/ instead of /:/.
 _
     args => {
         word  => { schema=>[str=>{default=>''}], pos=>0 },
+        ci => { schema=>'bool' },
     },
     result_naked => 1,
     result => {
@@ -145,6 +151,10 @@ $SPEC{complete_file} = {
             schema  => [str=>{default=>''}],
             pos     => 0,
         },
+        ci => {
+            summary => 'Case-insensitive matching',
+            schema  => 'bool',
+        },
         filter => {
             summary => 'Only return items matching this filter',
             description => <<'_',
@@ -174,8 +184,29 @@ _
 sub complete_file {
     my %args   = @_;
     my $word   = $args{word} // "";
+    my $ci     = $args{ci};
     my $filter = $args{filter};
 
+    # if word is starts with "~/" or "~foo/" replace it temporarily with user's
+    # name (so we can restore it back at the end). this is to mimic bash
+    # support. note that bash does not support case-insensitivity for "foo".
+    my $res_prefix; # to be added again to final result
+    my $search_prefix = ''; # to be added when searching
+    my $res_num_remove = 0;
+    if ($word =~ s!\A(~[^/]*)/!!) {
+        $res_prefix = "$1/";
+        my @dir = glob($1); # glob will expand ~foo to /home/foo
+        return [] unless @dir;
+        $search_prefix = $dir[0] =~ m!/\z! ? $dir[0] : "$dir[0]/";
+        $res_num_remove = length($search_prefix);
+    } elsif ($word =~ s!\A((?:\.\.?/+)+|/+)!!) {
+        $search_prefix = $1;
+        $search_prefix =~ s#/+\z## unless $search_prefix =~ m!\A/!;
+    } else {
+        $search_prefix = '';
+    }
+
+    # prepare filter sub
     if ($filter && !ref($filter)) {
         my @seqs = split /\s*\|\s*/, $filter;
         $filter = sub {
@@ -203,30 +234,91 @@ sub complete_file {
         };
     }
 
-    my @all;
-    if ($word =~ m!(\A|/)\z!) {
-        my $dir = length($word) ? $word : ".";
-        opendir my($dh), $dir or return [];
-        @all = map { ($dir eq '.' ? '' : $dir) . $_ }
-            grep { $_ ne '.' && $_ ne '..' } readdir($dh);
-        closedir $dh;
-    } else {
-        # must add wildcard char, glob() is convoluted. also {a,b} is
-        # interpreted by glob() (not so by bash file completion). also
-        # whitespace is interpreted by glob :(. later should replace with a
-        # saner one, like wildcard2re.
-        @all = glob("$word*");
+    # split word by dirs, as we want to dig level by level (needed when doing
+    # case-insensitive search on a case-sensitive fs). XXX we should optimize
+    # and not split .. or .
+    my @intermediate_dirs = split m!/+!, $word;
+    @intermediate_dirs = ('') if !@intermediate_dirs;
+    push @intermediate_dirs, '' if $word =~ m!/\z!;
+
+    # extract leaf path, because this one is treated differently
+    my $leaf = pop @intermediate_dirs;
+    @intermediate_dirs = ('') if !@intermediate_dirs;
+
+    #say "D:search_prefix=<$search_prefix>";
+    #say "D:intermediate_dirs=[",join(", ", map{"<$_>"} @intermediate_dirs),"]";
+    #say "D:leaf=<$leaf>";
+
+    # candidate for intermediate paths. when doing case-insensitive search,
+    # there maybe multiple candidate paths for each dir, for example if
+    # word='../foo/s' and there is '../foo/Surya', '../Foo/sri', '../FOO/SUPER'
+    # then candidate paths would be ['../foo', '../Foo', '../FOO'] and the
+    # filename should be searched inside all those dirs. everytime we drill down
+    # to deeper subdirectories, we adjust this list.
+    my @candidate_paths;
+
+    for my $i (0..$#intermediate_dirs) {
+        my $intdir = $intermediate_dirs[$i];
+        my @dirs;
+        if ($i == 0) {
+            # first path elem, we search search_prefix first since
+            # candidate_paths is still empty.
+            @dirs = ($search_prefix);
+        } else {
+            # subsequent path elem, we search all candidate_paths
+            @dirs = @candidate_paths;
+        }
+
+        if ($i == $#intermediate_dirs && $intdir eq '') {
+            @candidate_paths = @dirs;
+            last;
+        }
+
+        my @new_candidate_paths;
+        for my $dir (@dirs) {
+            #say "D:  intdir opendir($dir)";
+            opendir my($dh), ($dir eq '' ? '.' : $dir) or next;
+            # check if the deeper level is a directory with the expected name
+            my $re = $ci ? qr/\A\Q$intdir\E\z/i : qr/\A\Q$intdir\E\z/;
+            #say "D:  re=$re";
+            for (sort readdir $dh) {
+                #say "D:  $_";
+                next unless $_ =~ $re;
+                # skip . and .. if leaf is empty, like in bash
+                next if ($_ eq '.' || $_ eq '..') && $intdir eq '';
+                my $p = $dir =~ m!\A\z|/\z! ? "$dir$_" : "$dir/$_";
+                next unless -d $p;
+                push @new_candidate_paths, $p;
+            }
+        }
+        #say "D:  candidate_paths=[",join(", ", map{"<$_>"} @new_candidate_paths),"]";
+        return [] unless @new_candidate_paths;
+        @candidate_paths = @new_candidate_paths;
     }
 
-    my @words;
-    for (@all) {
-        next if $filter && !$filter->($_);
-        $_ = "$_/" if (-d $_) && !m!/\z!; # XXX double stat
-        #s!.+/(.+)!$1!;
-        push @words, $_;
+    my @res;
+    for my $dir (@candidate_paths) {
+        #say "D:opendir($dir)";
+        opendir my($dh), ($dir eq '' ? '.' : $dir) or next;
+        my $re = $ci ? qr/\A\Q$leaf/i : qr/\A\Q$leaf/;
+        #say "D:re=$re";
+        for (sort readdir $dh) {
+            next unless $_ =~ $re;
+            # skip . and .. if leaf is empty, like in bash
+            next if ($_ eq '.' || $_ eq '..') && $leaf eq '';
+            my $p = $dir =~ m!\A\z|/\z! ? "$dir$_" : "$dir/$_";
+            next if $filter && !$filter->($p);
+
+            # process into final result
+            substr($p, 0, $res_num_remove) = '' if $res_num_remove;
+            $p = "$res_prefix$p" if length($res_prefix);
+            $p .= "/" if -d $p;
+
+            push @res, $p;
+        }
     }
 
-    my $w = complete_array_elem(array=>\@words);
+    \@res;
 }
 
 # TODO: complete_filesystem (probably in a separate module)
@@ -248,7 +340,7 @@ Complete::Util - General completion routines
 
 =head1 VERSION
 
-This document describes version 0.13 of Complete::Util (from Perl distribution Complete-Util), released on 2014-07-13.
+This document describes version 0.14 of Complete::Util (from Perl distribution Complete-Util), released on 2014-12-05.
 
 =head1 DESCRIPTION
 
@@ -259,6 +351,8 @@ This document describes version 0.13 of Complete::Util (from Perl distribution C
 
 Complete from array.
 
+Will sort the resulting completion list, so you don't have to presort the array.
+
 Arguments ('*' denotes required arguments):
 
 =over 4
@@ -273,6 +367,8 @@ Arguments ('*' denotes required arguments):
 
 Return value:
 
+ (array)
+
 
 =head2 complete_env(%args) -> array
 
@@ -293,6 +389,8 @@ Arguments ('*' denotes required arguments):
 
 Return value:
 
+ (array)
+
 
 =head2 complete_file(%args) -> array
 
@@ -302,6 +400,10 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
+=item * B<ci> => I<bool>
+
+Case-insensitive matching.
+
 =item * B<filter> => I<code|str>
 
 Only return items matching this filter.
@@ -325,6 +427,8 @@ included.
 
 Return value:
 
+ (array)
+
 
 =head2 complete_hash_key(%args) -> array
 
@@ -344,6 +448,8 @@ Arguments ('*' denotes required arguments):
 
 Return value:
 
+ (array)
+
 
 =head2 complete_program(%args) -> array
 
@@ -355,12 +461,16 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
+=item * B<ci> => I<bool>
+
 =item * B<word> => I<str> (default: "")
 
 =back
 
 Return value:
 
+ (array)
+
 =for Pod::Coverage ^(complete_array)$
 
 =head1 SEE ALSO
@@ -368,7 +478,7 @@ Return value:
 L<Complete>
 
 If you want to do bash tab completion with Perl, take a look at
-L<Complete::Bash> and L<Complete::BashGen>.
+L<Complete::Bash> or L<Getopt::Long::Complete> or L<Perinci::CmdLine>.
 
 Other C<Complete::*> modules.
 
@@ -378,7 +488,7 @@ Please visit the project's homepage at L<https://metacpan.org/release/Complete-U
 
 =head1 SOURCE
 
-Source repository is at L<https://github.com/sharyanto/perl-SHARYANTO-Complete-Util>.
+Source repository is at L<https://github.com/perlancar/perl-Complete-Util>.
 
 =head1 BUGS
 
@@ -390,11 +500,11 @@ feature.
 
 =head1 AUTHOR
 
-Steven Haryanto <stevenharyanto@gmail.com>
+perlancar <perlancar@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2014 by Steven Haryanto.
+This software is copyright (c) 2014 by perlancar@cpan.org.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
